//
// CurveTable Envelope (modified)
// thanks Mike Clemow
//
class Globals
{
    0.01 => float TMIN;
    1. => float TMAX;
    40. => float PITCH_MIN;
    14000. => float PITCH_MAX;
    0.001 => float MIN_DURATION;
    0.01 => float MAX_DURATION;
    
    //again
    0. => float SEL_VOL;
    0. => float DISP_VOL; 
    (TMAX + TMIN)/2. => float SEL_TIME;
    0. => float DISP_TIME;  
    (PITCH_MAX + PITCH_MIN)/2. => float SEL_PITCH;
    (MAX_DURATION + MIN_DURATION)/2. => float DISP_PITCH;
    0.01 => float DURATION;
    2 => int ENV_TYPE;
}

class ListeningOSC
{
    OscRecv recv; 
    6449 => recv.port; 
    recv.listen(); 
    recv.event( "/instrument, i i i f" ) @=> OscEvent oe;
    Globals g;
    int ch; 
    
    fun void setChannel(int channel)
    {
        channel => ch;
    }
    
    fun void setGlobal(Globals glob)
    {
        glob @=> g;
    }
        
    fun void listen()
    {
        while ( true ) 
        { 
            oe => now; 
            while ( oe.nextMsg() != 0 ) 
            { 
                if(oe.getInt() == ch)
                {
                    oe.getInt() => float row; 
                    oe.getInt()   => int col_ch; 
                    oe.getFloat() => float value;
                    
                    <<< value + "=> [" + row + "," + col_ch + "] @ ch " + ch >>>; 
                    
                    if(row == 0 && col_ch == 0)
                    {
                        value => g.SEL_VOL;
                    }
                    else if(row == 0 && col_ch == 1)
                    {
                        value => g.DISP_VOL;
                    }
                    else if(row == 1 && col_ch == 0)
                    {
                        (g.TMAX - g.TMIN)*value + g.TMIN => g.SEL_TIME;
                    }
                    else if(row == 1 && col_ch == 1)
                    {
                        value => g.DISP_TIME;
                    }
                    else if(row == 2 && col_ch == 0)
                    {
                        (g.PITCH_MAX - g.PITCH_MIN)*value + g.PITCH_MIN => g.SEL_PITCH;
                    }
                    else if(row == 2 && col_ch == 1)
                    {
                        value => g.DISP_PITCH;
                    }
                    else if(row == 3 && col_ch == 0)
                    {
                        (g.MAX_DURATION - g.MIN_DURATION)*value + g.MIN_DURATION => g.DURATION;
                    }
                    else if(row == 3 && col_ch == 1)
                    {
                        if(value <= 0.33)
                        {
                            1 => g.ENV_TYPE;
                        }
                        else if(value <= 0.66)
                        {
                            2 => g.ENV_TYPE;;
                        }
                        else 3 => g.ENV_TYPE;
                    }

                    else if(row == 4 && col_ch == 0)
                    {
                        value => g.MIN_DURATION;
                    }
                    else if(row == 4 && col_ch == 1)
                    {
                        value => g.MAX_DURATION;
                    }
                    else if(row == 5 && col_ch == 0)
                    {
                        value => g.TMIN;
                    }
                    else if(row == 5 && col_ch == 1)
                    {
                        value => g.TMAX;
                    }
                    else if(row == 6 && col_ch == 0)
                    {
                        value => g.PITCH_MIN;
                    }                    
                    else if(row == 6 && col_ch == 1)
                    {
                        value => g.PITCH_MAX;
                    }
                }
            } 
        }
    } 
}

class CurveEnvelope
{
    // members and default params
    Phasor drive => CurveTable curvetable => Gain multiplier;
    3 => multiplier.op;				// multiply the curve to the signal to make an envelope
    0. => multiplier.gain;			// we use this to gait the output, so start at 0.
    
    UGen source, out;
    
    -1 => drive.op;					// stop the driver for now
    -1 => multiplier.op;
    
    //dac @=> out;
    1. => float gain;
    1::second => dur length;
    [0., 0., 0.,  1., 1., 0.,  2., 0.] => curvetable.coefs;	// triangle window (the sky is the limit!)
    
    // methods
    fun void setEnvelopeCoefs( float _coefs[] )
    {
        _coefs => curvetable.coefs;
    }
    
    fun void connectSource( UGen src )
    {
        src @=> source;
    }
    
    fun void connectOutput( UGen destination )
    {
        destination @=> out;
    }
    
    fun void triangular_env()
    {
        [0., 0., 0.,  1., 1., 0.,  2., 0.] => curvetable.coefs;  // triangle (default)
    }
    
    fun void percusive_env()
    {
        [0., 0., 2.,  0.1, 1., -5.,  1., 0.] => curvetable.coefs;  // percussive (like Env.perc in SuperCollider)
    }
    
    fun void rexpodec_env()
    {
        [0., 0., 5.,  0.9, 1., -2.,  1., 0.] => curvetable.coefs;  // reversed-percussive (Roads calls this type "rexpodec")
    }
    
    fun void trigger()
    {
        source => multiplier => out;			// connect things
        gain => multiplier.gain;				// open gait
        1 => drive.op;
        3 => multiplier.op;
        0. => drive.phase;						// reset driver to beginning of envelope curve
        1. / (length / second) => drive.freq;	// calculate speed of driver in Hz
        length => now;							// let it happen...
        0. => multiplier.gain;					// close gait
        -1 => drive.op;
        -1 => multiplier.op;
        source =< multiplier =< out;			// disconnect things
        me.yield();
    }
}

class MakeGrain
{
    10 => int n;
    0 => int index;
    0 => int cont;
    1. => float MASTER_GAIN;
    0 => int _id;
        
    SinOsc s[n];
    CurveEnvelope ce[n];
    Gain vol;
    Gain master;
    vol => master => dac;
    
    fun void setID(int id)
    {
        id => _id;
    }
    
    fun void grainLayer(float sel_vol, float vol_disp, float pitch, float pitch_disp, float duration, int env_type)
    {
        Std.rand2f( sel_vol - vol_disp, sel_vol + vol_disp) => vol.gain;
        sel_vol => vol.gain;
        
        1. => master.gain;
        
        Std.rand2f( pitch*(1 - pitch_disp), pitch*(1 + pitch_disp)) => s[index].freq;
       
        duration::second => ce[index].length;
        
        ce[index].connectSource( s[index] );
        ce[index].connectOutput( vol );
        
        if(env_type == 1)
        {
            ce[index].triangular_env();
        }
        else if(env_type  == 2)
        {
            ce[index].percusive_env();
        }
        else if(env_type  == 3)
        {
            ce[index].rexpodec_env();
        }
        
        spork ~ ce[index].trigger();
        me.yield();
        
        cont++;
        (duration/2.)::second => now;
        (index+1)%n => index;
    }   
}        

class Voice
{
    int NUMB;
    
    Globals g;
    GrainSync gs;
    ListeningOSC oscm;
    
    gs.setGlobal(g);
    oscm.setGlobal(g);
    
    fun void setVoiceNumb(int numb)
    {
        numb => NUMB;
        oscm.setChannel(NUMB);
    }
    
    fun void start()
    {
        spork ~ oscm.listen();
        spork ~ gs.metro();
    }
}

class GrainSync
{
    MakeGrain mg;
    Globals g;
    
    0 => int metro_type;

    Event e;
    
    fun void setGlobal(Globals glob)
    {
        glob @=> g;
    }
    
    fun void makeIt()
    {
        mg.grainLayer(g.SEL_VOL, g.DISP_VOL, g.SEL_PITCH, g.DISP_PITCH, g.DURATION, g.ENV_TYPE);
    }
    
    fun void setMetroType()
    {
        if(metro_type == 0)
        {
            1 => metro_type;
        }
        else
        {
            0 => metro_type;
            e.signal();
        }
    }
    
    fun void metro()
    {
        while( true )
        {
            if(metro_type == 0)
            {
                makeIt();
                now + currentTime()::second => now;
            }    
            else
            {
                e => now;
            }
        }
    }
    
    fun float currentTime()
    {
        if(g.DISP_TIME != 0)
        {
            return Std.rand2f( g.SEL_TIME*( 1 - g.DISP_TIME ), g.SEL_TIME*( 1 + g.DISP_TIME  ));
        }
        else
        {
            return g.SEL_TIME;
        }
    }
}

//MAIN
for(0 => int i; i < 8; i++)
{
    Voice v;
    v.setVoiceNumb(i);
    v.start(); 
}

while( true )
    1::day => now;
